(******************************************************************************)
(* generator (stack machine with condition register)                          *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Gen;

FROM SYSTEM IMPORT ADR, BYTE;
IMPORT Lib, FIO;
FROM Scnr IMPORT TSymbol;
FROM Comp0 IMPORT TOperation, TFunction, TCmdDesc, TSet, TCmdStr;
FROM Comp0 IMPORT aNone, acMax, vcMax, pcMax, Convert;
FROM Comp IMPORT Write, GetPos, SetPos, Read, Trunc;

CONST
  debug = TRUE;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

VAR
  cmd: TCmdDesc;
  (* debug *)
  first: BOOLEAN;
  (* constants *)
  vals: ARRAY BOOLEAN OF TValue;


PROCEDURE DbgCmd();
VAR a: TAddress; s: TCmdStr;
BEGIN
  GetPos( a );
  Convert( cmd, s );
  IF first THEN FIO.WrStr( dbg, 'Gen.Dbg:  pos  cmd' ); FIO.WrLn( dbg ) END;
  first := FALSE;
  FIO.WrStr( dbg, 'Gen.Dbg: ' );
  FIO.WrCard( dbg, a, 4 );
  FIO.WrStr( dbg, ': ' );
  FIO.WrStr( dbg, s );
  FIO.WrLn( dbg );
END DbgCmd;

PROCEDURE DbgFix( a, v: TAddress );
BEGIN
  FIO.WrStr( dbg, 'Gen.Dbg: Fixup ' );
  FIO.WrCard( dbg, a, 0 );
  FIO.WrStr( dbg, ' with ' );
  FIO.WrCard( dbg, v, 0 );
  FIO.WrLn( dbg );
END DbgFix;

PROCEDURE DbgSav();
BEGIN
  FIO.WrStr( dbg, 'Gen.Dbg: Save' );
  FIO.WrLn( dbg );
END DbgSav;

PROCEDURE DbgRes( a: TAddress );
BEGIN
  FIO.WrStr( dbg, 'Gen.Dbg: Restore ' );
  FIO.WrCard( dbg, a, 0 );
  FIO.WrLn( dbg );
END DbgRes;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Gen: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE ErrCard( s: ARRAY OF CHAR; x: CARDINAL );
BEGIN
  FIO.WrStr( err, 'Gen: ' );
  FIO.WrStr( err, s );
  FIO.WrChar( err, ' ' );
  FIO.WrCard( err, x, 0 );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END ErrCard;


PROCEDURE Hlt();
BEGIN
  cmd.fct := fHlt;
  IF debug THEN DbgCmd() END;
  Write( cmd.fct );
END Hlt;

PROCEDURE Val( v: TValue );
BEGIN
  cmd.fct := fVal; cmd.val := v;
  IF debug THEN DbgCmd() END;
  Write( cmd.fct ); Write( cmd.val );
END Val;

PROCEDURE Lod( i: TIndex );
BEGIN
  IF i >= vcMax THEN Err( 'Lod: Invalid index' ) END;
  cmd.fct := fLod; cmd.idx := i;
  IF debug THEN DbgCmd() END;
  Write( cmd.fct ); Write( cmd.idx );
END Lod;

PROCEDURE Sto( i: TIndex );
BEGIN
  IF i >= vcMax THEN Err( 'Sto: Invalid index' ) END;
  cmd.fct := fSto; cmd.idx := i;
  IF debug THEN DbgCmd() END;
  Write( cmd.fct ); Write( cmd.idx );
END Sto;

PROCEDURE Opr( o: TOperation );
BEGIN
  cmd.fct := fOpr; cmd.opr := o;
  IF debug THEN DbgCmd() END;
  Write( cmd.fct ); Write( cmd.opr );
END Opr;

PROCEDURE Jmp( VAR a: TAddress );
BEGIN
  cmd.fct := fJmp; cmd.adr := a;
  IF debug THEN DbgCmd() END;
  Write( cmd.fct ); GetPos( a ); Write( cmd.adr );
END Jmp;

PROCEDURE Jpc( VAR a: TAddress; c: BOOLEAN );
BEGIN
  IF c THEN cmd.fct := fJpt ELSE cmd.fct := fJpf END; cmd.adr := a;
  IF debug THEN DbgCmd() END;
  Write( cmd.fct ); GetPos( a ); Write( cmd.adr );
END Jpc;

PROCEDURE Save( VAR a: TAddress );
BEGIN
  IF debug THEN DbgSav() END;
  GetPos( a );
END Save;

PROCEDURE Restore( a: TAddress );
BEGIN
  IF debug THEN DbgRes( a ) END;
  SetPos( a ); Trunc();
END Restore;


PROCEDURE Init( VAR a: TAddress );
BEGIN
  a := aNone;
END Init;

PROCEDURE FixWith( v, a: TAddress );
VAR pos, nxt: TAddress;
BEGIN
  GetPos( pos );
  WHILE a # aNone DO
    IF debug THEN DbgFix( a, v ) END;
    IF a > pos THEN ErrCard( 'FixWith.Bug: Invalid address', a ) END;
    SetPos( a ); Read( nxt );
    SetPos( a ); Write( v );
    a := nxt;
  END;
  SetPos( pos );
END FixWith;

PROCEDURE Fixup( a: TAddress );
VAR pos: TAddress;
BEGIN
  GetPos( pos ); FixWith( pos, a );
END Fixup;

PROCEDURE Load( VAR x: TItemDesc );
VAR a,b: TAddress;
BEGIN
  CASE x.m OF
  | mConst: Val( x.v );
  | mVar  : Lod( x.i );
  | mStack:
  | mCond : Init( a ); Init( b );
            Jpc( a, ~x.c );
            Fixup( x.b ); Val( vTrue ); Jmp( b );
            Fixup( x.a ); Fixup( a ); Val( vFalse );
            Fixup( b );
  ELSE
    ErrCard( 'Load.Bug: Invalid mode', ORD(x.m) );
  END;
  x.m := mStack;
END Load;

PROCEDURE SetCond( VAR x: TItemDesc );
BEGIN
  x.m := mCond; x.c := TRUE; Init( x.a ); Init( x.b );
END SetCond;

PROCEDURE LoadCond( VAR x: TItemDesc );
BEGIN
  IF x.m = mCond THEN Err( 'LoadCond.Bug: Already in mCond' ) END;
  Load( x ); Val( vFalse ); Opr( oNeq ); SetCond( x );
END LoadCond;

PROCEDURE Wrt( x: TItemDesc );
BEGIN
  CASE x.t OF
  | tInt : Load( x ); Opr( oWrt );
  | tBool: Load( x ); Val( vFalse ); Opr( oNeq ); Opr( oWrC );
  ELSE
    ErrCard( 'Wrt.Bug: Invalid type', ORD(x.t) );
  END;
END Wrt;

PROCEDURE Op1( s: TSymbol; VAR x: TItemDesc );
VAR a: TAddress;
BEGIN
  IF x.m = mConst THEN
    CASE s OF
    | sMinus    : x.v := -x.v;
    | sNot      : x.v := vals[ x.v = vFalse ];
    | sAnd      : IF x.v = vFalse THEN Save( x.p ) END;
    | sOr       : IF x.v # vFalse THEN Save( x.p ) END;
    | sEql..sGeq:
    ELSE
      ErrCard( 'Op1.Calc.Bug: Invalid symbol', ORD(s) );
    END;
  ELSE
    CASE s OF
    | sMinus    : Load( x ); Opr( oNeg );
    | sNot      : IF x.m # mCond THEN LoadCond( x ) END;
                  x.c := ~x.c; a := x.a; x.a := x.b; x.b := a;
    | sAnd      : IF x.m # mCond THEN LoadCond( x ) END;
                  Jpc( x.a, ~x.c ); Fixup( x.b ); Init( x.b );
    | sOr       : IF x.m # mCond THEN LoadCond( x ) END;
                  Jpc( x.b, x.c ); Fixup( x.a ); Init( x.a );
    | sEql..sGeq: IF x.m = mCond THEN Load( x ) END;
    ELSE
      ErrCard( 'Op1.Gen.Bug: Invalid symbol', ORD(s) );
    END;
  END;
END Op1;

PROCEDURE Merge( VAR a: TAddress; b: TAddress );
VAR pos, ret, tmp: TAddress;
BEGIN
  IF b # aNone THEN
    ret := b; GetPos( pos );
    LOOP
      SetPos( b ); Read( tmp );
      IF tmp = aNone THEN EXIT END;
      b := tmp;
    END;
    SetPos( b ); Write( a );
    SetPos( pos ); a := ret;
  END;
END Merge;

PROCEDURE Op2( s: TSymbol; VAR x: TItemDesc; y: TItemDesc );
VAR o: TOperation;

  PROCEDURE SetOpr();
  BEGIN
    CASE s OF
    | sPlus : o := oAdd;
    | sMinus: o := oSub;
    | sMul  : o := oMul;
    | sDiv  : o := oDiv;
    | sOr   : o := oOr;   (* unused *)
    | sAnd  : o := oAnd;  (* unused *)
    | sEql  : o := oEql;
    | sNeq  : o := oNeq;
    | sLss  : o := oLss;
    | sLeq  : o := oLeq;
    | sGtr  : o := oGtr;
    | sGeq  : o := oGeq;
    ELSE
      ErrCard( 'Op2.SetOpr.Bug: Invalid symbol', ORD(s) );
    END;
  END SetOpr;

  PROCEDURE Calc();
  BEGIN
    CASE o OF
    | oAdd: x.v := x.v + y.v;
    | oSub: x.v := x.v - y.v;
    | oMul: x.v := x.v  *  y.v;
    | oDiv: x.v := x.v DIV y.v;
    | oOr : x.v := TValue( TSet(x.v) + TSet(y.v) );
    | oAnd: x.v := TValue( TSet(x.v) * TSet(y.v) );
    | oEql: x.v := vals[ x.v =  y.v ];
    | oNeq: x.v := vals[ x.v #  y.v ];
    | oLss: x.v := vals[ x.v <  y.v ];
    | oLeq: x.v := vals[ x.v <= y.v ];
    | oGtr: x.v := vals[ x.v >  y.v ];
    | oGeq: x.v := vals[ x.v >= y.v ];
    ELSE
      ErrCard( 'Op2.Calc.Bug: Invalid operation', ORD(o) );
    END;
  END Calc;

  PROCEDURE InvOpr();
  BEGIN
    CASE o OF
    | oAdd:
    | oSub: o := oBus;
    | oMul:
    | oDiv: o := oVid;
    | oEql: o := oNeq;
    | oNeq: o := oEql;
    | oLss: o := oGeq;
    | oLeq: o := oGtr;
    | oGtr: o := oLeq;
    | oGeq: o := oLss;
    ELSE
      ErrCard( 'Op2.InvOpr.Bug: Invalid operation', ORD(o) );
    END;
  END InvOpr;

BEGIN
  SetOpr();
  IF ( x.m = mConst ) & ( y.m = mConst ) THEN
    Calc();
  ELSE
    IF o = oAnd THEN
      IF x.m = mConst THEN
        IF x.v = vFalse THEN Restore( x.p ) ELSE x := y END;
      ELSE
        IF x.m # mCond THEN ErrCard( 'Op2.And.Bug: Invalid mode', ORD(x.m) ) END;
        IF y.m # mCond THEN LoadCond( y ) END;
        x.c := y.c; Merge( x.a, y.a ); x.b := y.b;
      END;
    ELSIF o = oOr THEN
      IF x.m = mConst THEN
        IF x.v # vFalse THEN Restore( x.p ) ELSE x := y END;
      ELSE
        IF x.m # mCond THEN ErrCard( 'Op2.Or.Bug: Invalid mode', ORD(x.m) ) END;
        IF y.m # mCond THEN LoadCond( y ) END;
        x.c := y.c; x.a := y.a; Merge( x.b, y.b );
      END;
    ELSE
      IF ( x.m # mStack ) & ( y.m = mStack ) THEN InvOpr() END;
      Load( x ); Load( y ); Opr( o );
      IF ( oEql <= o ) & ( o <= oGeq ) THEN SetCond( x ) END;
    END;
  END;
END Op2;

PROCEDURE Asg( x, y: TItemDesc );
BEGIN
  IF x.m # mVar THEN Err( 'Asg: Not a variable' ) END;
  Load( y ); Sto( x.i );
END Asg;

PROCEDURE Label( VAR a: TAddress );
BEGIN
  GetPos( a );
END Label;

PROCEDURE JmpFalse( VAR x: TItemDesc );
BEGIN
  IF x.m = mConst THEN
    IF x.v = vFalse THEN SetCond( x ); Jmp( x.a ) ELSE SetCond( x ) END;
    RETURN;
  END;
  IF x.m # mCond THEN LoadCond( x ) END;
  Jpc( x.a, ~x.c ); Fixup( x.b ); Init( x.b );
END JmpFalse;


PROCEDURE InitModule();
BEGIN
  first := TRUE;

  Lib.Fill( ADR(vals), SIZE(vals), 0C );
  vals[ FALSE ] := vFalse;
  vals[ TRUE  ] := vTrue;
END InitModule;

BEGIN
  InitModule();
END Gen.
